# Generated by Django 4.2.6 on 2023-10-10 16:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_csv", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="csvdownload",
            name="columns",
            field=models.TextField(
                help_text="The list of source columns included in the download",
            ),
        ),
    ]



from django.contrib import admin

from .models import CsvDownload


class CsvDownloadAdmin(admin.ModelAdmin):
    list_display = ("user", "timestamp", "row_count", "filename")
    list_filter = ("timestamp",)
    search_fields = ("user", "filename")
    raw_id_fields = ("user",)
    readonly_fields = (
        "user",
        "timestamp",
        "filename",
        "row_count",
        "columns",
    )


admin.site.register(CsvDownload, CsvDownloadAdmin)



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _lazy


class CsvDownload(models.Model):
    """Track CSV downloads."""

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        on_delete=models.SET_NULL,
        help_text="User who initiated the download.",
    )
    filename = models.CharField(max_length=100)
    timestamp = models.DateTimeField(
        auto_now_add=True, help_text=_lazy("When the download took place.")
    )
    row_count = models.IntegerField(
        null=True, blank=True, help_text=_lazy("Rows downloaded")
    )
    columns = models.TextField(
        help_text=_lazy("The list of source columns included in the download"),
    )

    class Meta:
        verbose_name = "CSV Download"

    def __str__(self) -> str:
        return f"{self.filename}"



from typing import Any, List, Type

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models.query import QuerySet
from django.http import HttpRequest, HttpResponse
from django.views import View

from .csv import BaseQuerySetWriter, BulkQuerySetWriter, write_csv
from .models import CsvDownload
from .settings import MAX_ROWS
from .types import OptionalSequence


def download_csv(
    user: settings.AUTH_USER_MODEL,
    filename: str,
    queryset: QuerySet,
    *columns: str,
    header: bool = True,
    max_rows: int = MAX_ROWS,
    column_headers: OptionalSequence = None,
    writer_klass: Type[BaseQuerySetWriter] = BulkQuerySetWriter,
    **writer_kwargs: Any,
) -> HttpResponse:
    """Download queryset as a CSV."""
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    row_count = write_csv(
        response,
        queryset,
        *columns,
        header=header,
        max_rows=max_rows,
        column_headers=column_headers,
        writer_klass=writer_klass,
        **writer_kwargs,
    )
    response["X-Row-Count"] = row_count
    CsvDownload.objects.create(
        user=user,
        row_count=row_count,
        filename=filename,
        columns=", ".join(columns),
    )
    return response


class CsvDownloadView(View):
    """CBV for downloading CSVs."""

    writer_klass = BulkQuerySetWriter

    def get_writer_klass(self) -> Type[BaseQuerySetWriter]:
        # Override to provide a different writer
        return self.writer_klass

    def get_writer_kwargs(self) -> dict:
        # custom kwargs for initialising the writer
        return {}

    def has_permission(self, request: HttpRequest) -> bool:
        """Return True if the user has permission to download this file."""
        return True

    def get_max_rows(self, request: HttpRequest) -> int:
        """Override to set custom MAX_ROWS on a per-request basis."""
        return MAX_ROWS

    def add_header(self, request: HttpRequest) -> bool:
        """Return True to include header row in CSV."""
        return True

    def get_user(self, request: HttpRequest) -> settings.AUTH_USER_MODEL:
        """
        Return the user against whom to record the download.

        This is provided for cases where the request.user may not be the
        user you wish to record the download against. Impersonation is the
        canonical use case for overriding this.

        """
        return request.user

    def get_filename(self, request: HttpRequest) -> str:
        """Return download filename."""
        raise NotImplementedError

    def get_columns(self, request: HttpRequest) -> List[str]:
        """Return columns to extract from the queryset."""
        raise NotImplementedError

    def get_column_headers(self, request: HttpRequest) -> List[str]:
        """Return column headers to apply to the CSV."""
        return self.get_columns(request)

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        """Return the data to be downloaded."""
        raise NotImplementedError

    def get(self, request: HttpRequest) -> HttpResponse:
        """Download data as CSV."""
        if not self.has_permission(request):
            raise PermissionDenied

        return download_csv(
            self.get_user(request),
            self.get_filename(request),
            self.get_queryset(request),
            *self.get_columns(request),
            header=self.add_header(request),
            max_rows=self.get_max_rows(request),
            column_headers=self.get_column_headers(request),
            writer_klass=self.get_writer_klass(),
            **self.get_writer_kwargs(),
        )



"""
Functions used to write QuerySet data to a CSV.

This module contains a class (QuerySetWriter) used to wrap up writing of
data from a QuerySet to a CSV (file or HttpResponse).

You should never need to call the QuerySetWriter explicitly - but instead
use the `write_csv` function.

Example of writing to a file:

    >>> qs = MyModel.objects.all()
    >>> cols = ("col1", "col2", "col3")
    >>> with open('foo.csv', 'w') as csvfile:
    >>>     csv.write_csv(csvfile, qs, *cols)
    10

Example of writing to an HttpResponse:

    >>> response = HttpResponse(content_type="text/csv")
    >>> response["Content-Disposition"] = 'attachment; filename="foo.csv"'
    >>> csv.write_csv(response, qs, *cols)
    10

Example of writing to an in-memory text buffer:

    >>> buffer = io.StringIO()
    >>> csv.write_csv(buffer, qs, *cols)
    10

"""

import csv
import logging
from typing import Any, Sequence, Type

from django.core.paginator import Paginator
from django.db.models import QuerySet

from .settings import DEFAULT_PAGE_SIZE, MAX_ROWS
from .types import OptionalSequence

logger = logging.getLogger(__name__)


class BaseQuerySetWriter:
    """
    Class used to write contents of a QuerySet to a CSV.

    Do not use this class directly - use one of the subclasses, based
    on your expected use case - data size, memory constraints etc.

    This class wraps the csv.writerow and csv.writerows functions, and
    maps queryset columns to CSV fields (via `values_list`). It is
    initialised with a 'csvfile' where "csvfile can be any object with
    a write() method."

    See https://docs.python.org/3/library/csv.html#csv.writer

    """

    def __init__(
        self, csvfile: Any, queryset: QuerySet, *columns: str, max_rows: int = MAX_ROWS
    ) -> None:
        self.writer = csv.writer(csvfile)
        self.queryset = queryset
        self.columns = columns
        self.max_rows = max_rows

    def rows(self) -> QuerySet:
        """Return the rows to write as a capped values_list queryset."""
        return self.queryset.values_list(*self.columns)[: self.max_rows]

    def write_header(self, column_headers: OptionalSequence = None) -> None:
        row: Sequence = self.columns
        if column_headers:
            if len(column_headers) != len(self.columns):
                raise ValueError("Columns and headers do not match in length.")
            row = column_headers
        self.writer.writerow(row)

    def write_rows(self) -> int:
        raise NotImplementedError


class BulkQuerySetWriter(BaseQuerySetWriter):
    """Subclass of QuerySetWriter that writes out queryset in one go."""

    def write_rows(self) -> int:
        """Write the rows out in one go."""
        self.writer.writerows(rows := self.rows())
        return rows.count()


class PagedQuerySetWriter(BaseQuerySetWriter):
    """Subclass of QuerySetWriter that writes out queryset in pages."""

    def __init__(self, *args: Any, page_size: int = DEFAULT_PAGE_SIZE, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.page_size = page_size

    def write_rows(self) -> int:
        """Write the rows out in pages."""
        paginator = Paginator(rows := self.rows(), self.page_size)
        for page_number in paginator.page_range:
            self.writer.writerows(paginator.page(page_number).object_list)
        return rows.count()


class RowQuerySetWriter(BaseQuerySetWriter):
    """Subclass of QuerySetWriter that writes out queryset row-by-row."""

    def write_rows(self) -> int:
        """Write the rows out one-by-one."""
        # Since using an iterator means the querysets result-cache is not populated,
        # a call to .count() will cause a new database hit, which can be very expensive
        # for large querysets. It's more efficient to use a manual counter here.
        row_count = 0
        for row in self.rows().iterator():
            self.writer.writerow(row)
            row_count += 1
        return row_count


def write_csv(
    fileobj: Any,
    queryset: QuerySet,
    *columns: str,
    header: bool = True,
    max_rows: int = MAX_ROWS,
    column_headers: OptionalSequence = None,
    writer_klass: Type[BaseQuerySetWriter] = BulkQuerySetWriter,
    **writer_kwargs: Any,
) -> int:
    """Write QuerySet to fileobj in CSV format using BulkQuerySetWriter."""
    writer = writer_klass(
        fileobj, queryset, *columns, max_rows=max_rows, **writer_kwargs
    )
    if header:
        writer.write_header(column_headers=column_headers)
    return writer.write_rows()
